package com.swop;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;

/**
 * <p>
 *  Generates PLSQL code for translating a given a html-string. 
 *  The string may be complex
 *  and use translation-includes and substitution variables. Follow the direction
 *  contained in the resulting string.
 * </p>
 * <p>
 *  Usage (see main-method):
 *  <blockquote>
 *          String convertedText=new CodeGenerator().generatePLSQLCode(rawText);
 *          System.out.println(convertedText);  
 *  </blockquote>
 * </p>
 * <p>
 *  If you use includes, the generator does not expect the keys but the full 
 *  include definition. Use the following syntax (note the " : "):
 *  <blockquote>
 *      rawText = "any text +#INCLUDE_KEY:the include text in any language +# more text";
 *  </blockquote>
 *  The text may contain substitution variables. You just need to use the keys:
 *  <blockquote>
 *      rawText = "any text #MY_VARIABLE# more text";
 *  </blockquote>
 *  </p>
 *  <p>
 *  The generator will generate new a constant for the given include-keys and for the
 *  substitution variables, and 
 *  create "create_translation"-statements (using the constant and the given include text). 
 *  Also in the resulting
 *  PLSQL-code, the includes and the variables are being replaced in the original rawText 
 *  by their PLSQL-constants. Nested includes are not being supported. Variables may be
 *  defined inside an include.
 * </p>
 * 
 * @see #setAddVariablesAsPlaceHolder(boolean addVariablesAsPlaceHolder)
 * @see #setUseIncludeSyntax(boolean useIncludeSyntax)
 * 
 */
public class CodeGenerator {
    public static final String INCLUDE_DELIMITER="+#";
    public static final String TEMP_INCLUDE_DELIMITER="@";
    
    public CodeGenerator() {
    }
    
    private static String decodeToTemp(String rawText) {
        return rawText.replace(INCLUDE_DELIMITER, TEMP_INCLUDE_DELIMITER);
    }

    private static String encodeFromTemp(String tempText) {
        return tempText.replace(TEMP_INCLUDE_DELIMITER, INCLUDE_DELIMITER);
    }
    
    private Map<String, String> includeMap=new HashMap<String, String>();
    private Map<String, String> includeConstantMap=new HashMap<String, String>();
    private Map<String, String> variableConstantMap=new HashMap<String, String>();
    
    private String convertedText=null;
    private String rawText=null;
    
    /**
     * if true, variables are added using the plsqholder-syntax and replace methods,
     * else just plain PLSQL-variable concatination are used
     */
    private boolean addVariablesAsPlaceHolder=false;
    
    /**
     * whether the include-syntax should be used in the resulting text or just plain
     * PLSQL-variable concatination
     */
    private boolean useIncludeSyntax=false;
    
    /**
     * compiles the given HTML-string in the syntax described above. Returns
     * any error
     * 
     * @param rawText  a HTML-string using the syntax for includes and subsituation-variables
     * @return true if an error occurred
     */
    private boolean compile(String rawText) {    
        boolean returnError=false;
        
        this.rawText = decodeToTemp(rawText);

        
        this.convertedText="";
        if(this.rawText!=null) {
            if(this.rawText.indexOf("=\"#\"")!=-1) {
                returnError=true;
                log("Please remove any self links href=\"#\"");
            }
                
            StringTokenizer includeTokens=new StringTokenizer(this.rawText, TEMP_INCLUDE_DELIMITER);
            while(includeTokens.hasMoreTokens()) {
                // get every second token as this will be whats inside:
                // outside +# inside +# outside +# inside +#
                String aInclude=null;
                String aBeforeInclude=includeTokens.nextToken();
                if(includeTokens.hasMoreTokens())
                    aInclude=includeTokens.nextToken();
                else
                    aInclude=null;
                
                if(aInclude!=null) {
                    StringTokenizer includeParts=new StringTokenizer(aInclude, ":");
                    String includeKey=null;
                    String includeValue=null;
                    String includeConstant=null;
                    if(includeParts.countTokens()!=2) {
                        returnError=true;
                        log("the following include requires the syntax <key:text>: <"+aInclude+">");
                        includeKey=aInclude;
                        includeValue=aInclude;
                        includeConstant=null;
                    } else {
                        includeKey=includeParts.nextToken();
                        includeValue=includeParts.nextToken();
                        int includeKeyLength=includeKey.length();
                        if(includeKeyLength>25)
                            includeConstant="G_TK_"+includeKey.toUpperCase().substring(0, 25);
                        else
                            includeConstant="G_TK_"+includeKey.toUpperCase();
                    }
                    
                    this.includeMap.put(includeKey, includeValue);        
                    this.includeConstantMap.put(includeKey, includeConstant);
                    
                    String includeStr="";
                    if(this.useIncludeSyntax)
                        includeStr=TEMP_INCLUDE_DELIMITER+"'||pkg_translation."+includeConstant+"||'"+TEMP_INCLUDE_DELIMITER;
                    else
                        includeStr="'||l_"+includeKey.toLowerCase()+"||'";
                        
                    this.convertedText=this.convertedText+aBeforeInclude+includeStr;
                } else {
                    // no include, the "before-part" is the text after the last include, add it:
                    this.convertedText=this.convertedText+aBeforeInclude;
                }
            }
            
            StringTokenizer variableTokens=new StringTokenizer(this.rawText, "#");
            if(variableTokens!=null) {
                while(variableTokens.hasMoreTokens()) {
                    String preVariableText=variableTokens.nextToken();
                    if(variableTokens.hasMoreTokens()) {
                        String variableKey=variableTokens.nextToken();
                        String variableConstant=null;
                        int variableKeyLength=variableKey.length();
                        if(variableKeyLength>25)
                            variableConstant="G_PH_"+variableKey.toUpperCase().substring(0, 25);
                        else
                            variableConstant="G_PH_"+variableKey.toUpperCase();                        
                        this.variableConstantMap.put(variableKey, variableConstant);
                        
                        String replacementStr="pkg_translation."+variableConstant;
                        if(!this.addVariablesAsPlaceHolder)
                            replacementStr =  "l_"+variableKey.toLowerCase();
                        
                        this.convertedText=
                            this.convertedText.replace("#"+variableKey+"#", 
                                                    "'||"+replacementStr+"||'");
                    }
                }
            }            
        }
        
        return returnError;
    }
    
    /**
     * Generates the PLSQL-code after the HTML-text was compiled
     * 
     * @return
     */
    private String outputCode() {
        String returnStr="";
        boolean error=false;
        if(!error) {
            // add the create-statements for the install-scripts:
            returnStr = returnStr + "\n\n Paste and adapt this code in your pkg_translation package:\n";

            List<String> variableKeyColl=new LinkedList<String>(this.variableConstantMap.keySet());
            Collections.sort(variableKeyColl);
            if(variableKeyColl!=null) {
                for(String aVariableKey : variableKeyColl) {
                    String aVariableConstant=(String) this.variableConstantMap.get(aVariableKey);
            
                    String createTransStr=
                        aVariableConstant+" CONSTANT VARCHAR2(50) :='#"+aVariableKey+"#';\n";
                    returnStr = returnStr+createTransStr;
                }
            }
            returnStr = returnStr + "\n";
            List<String> includeKeyColl=new LinkedList<String>(this.includeMap.keySet());
            Collections.sort(includeKeyColl);
            if(includeKeyColl!=null) {
                for(String aIncludeKey : includeKeyColl) {
                    String aIncludeConstant=(String) this.includeConstantMap.get(aIncludeKey);
            
                    String createTransStr=
                        aIncludeConstant+" CONSTANT VARCHAR2(50) :='"+aIncludeKey+"';\n";
                    returnStr = returnStr+createTransStr;
                }
            }
            
            returnStr = returnStr + "\n\n Paste and adapt this code in your installation scripts \n";
            
            if(includeKeyColl!=null) {
                for(String aIncludeKey : includeKeyColl) {
                    String aIncludeStr=(String) this.includeMap.get(aIncludeKey);
                    String aIncludeConstant= (String) this.includeConstantMap.get(aIncludeKey);
                    String createTransStr=
                        "pkg_translation.create_translation('de', pkg_translation."+aIncludeConstant+", '"+aIncludeStr+"');\n"+
                        "pkg_translation.create_translation('en', pkg_translation."+aIncludeConstant+", '"+aIncludeStr+"');\n";
                
                    returnStr = returnStr+createTransStr;
                }
            }
            
            // variables in the runtime code
            if(!error) {
                returnStr = returnStr + "\n\n Paste and adapt this code in your runtime-script:\n\n";
                if(!isUseIncludeSyntax() && this.includeConstantMap.size()>0) {
                    List<String> includeKeyList=new LinkedList<String>(this.includeConstantMap.keySet());
                    Collections.sort(includeKeyList);
                    for(String aVariable : includeKeyList) {
                        String aVariableConstant=this.includeConstantMap.get(aVariable);
                        if(aVariable.length()>28)
                            aVariable="l_"+aVariable.toLowerCase().substring(0, 28);
                        else
                            aVariable   = "l_"+aVariable.toLowerCase();
                        returnStr=returnStr+"   "+aVariable+" VARCHAR2(200); \n";
                    }
                }   
                if(this.variableConstantMap.size()>0) {
                    List<String> variableKeyList=new LinkedList<String>(this.variableConstantMap.keySet());
                    Collections.sort(variableKeyList);
                    for(String aVariable : variableKeyList) {
                        String aVariableConstant=this.variableConstantMap.get(aVariable);
                        if(aVariable.length()>28)
                            aVariable="l_"+aVariable.toLowerCase().substring(0, 28);
                        else
                            aVariable   = "l_"+aVariable.toLowerCase();
                        returnStr=returnStr+"   "+aVariable+" VARCHAR2(200); \n";
                    }
                }   
                
                returnStr=returnStr+"\n";
                returnStr=returnStr+"   l_text VARCHAR2(32767); \n";
                returnStr=returnStr+"   l_language VARCHAR2(2) :='de';\n";
                
                returnStr = returnStr + "\n";
                
                // if not include syntax: initialize the plsql-variables for the includes
                // using pkg_translation.get_translation
                if(!isUseIncludeSyntax() && this.includeConstantMap.size()>0) {
                    List<String> includeKeyList=new LinkedList<String>(this.includeConstantMap.keySet());
                    Collections.sort(includeKeyList);
                    for(String aVariable : includeKeyList) {
                        String aVariableConstant=this.includeConstantMap.get(aVariable);
                        if(aVariable.length()>28)
                            aVariable="l_"+aVariable.toLowerCase().substring(0, 28);
                        else
                            aVariable   = "l_"+aVariable.toLowerCase();
                        returnStr=returnStr+"   "+aVariable+" := pkg_translation.get_translation(p_lang, pkg_translation."+aVariableConstant+");\n";
                    }
                }                   
                returnStr = returnStr + "l_text := '"+this.convertedText+"';\n\n";
                returnStr = returnStr + "l_text := pkg_translation.get_translation(p_lang, l_text);\n\n";
                            
                // add replace calls for the placeholder variables:
                if(this.variableConstantMap.size()>0) {
                    List<String> variableKeyList=new LinkedList<String>(this.variableConstantMap.keySet());
                    Collections.sort(variableKeyList);
                    
                    for(String aVariable : variableKeyList) {
                        String aVariableConstant=this.variableConstantMap.get(aVariable);
                        returnStr=returnStr+"l_text := replace(l_text, pkg_translation."+aVariableConstant+", l_"+aVariable.toLowerCase()+");\n";
                    }
                }
            }
            returnStr=encodeFromTemp(returnStr);
        }
        
        return returnStr;
    }
    
    /**
     * compiles the HTML-string and returns the PLSQL-code
     * 
     * @param rawText  a HTML-string using the syntax for includes and subsituation-variables
     * @return         the generate code
     */
    public String generatePLSQLCode(String rawText) { 
        String returnStr=null;
        if(!compile(rawText))
            returnStr=outputCode();
        
        return returnStr;
    }
        

    /**
     * set to true, if variables are added using the plsqholder-syntax and replace methods,
     * else just plain PLSQL-variables are used
     */
    public void setAddVariablesAsPlaceHolder(boolean addVariablesAsPlaceHolder) {
        this.addVariablesAsPlaceHolder = addVariablesAsPlaceHolder;
    }

    /**
     * if true, variables are added using the plsqholder-syntax and replace methods,
     * else just plain PLSQL-variables are used
     */
    public boolean isAddVariablesAsPlaceHolder() {
        return addVariablesAsPlaceHolder;
    }

    /**
     * set whether the include-syntax should be used in the resulting text or just plain
     * PLSQL-variable concatination
     */
    public void setUseIncludeSyntax(boolean useIncludeSyntax) {
        this.useIncludeSyntax = useIncludeSyntax;
    }
    
    /**
     * gets whether the include-syntax should be used in the resulting text or just plain
     * PLSQL-variable concatination
     */
    public boolean isUseIncludeSyntax() {
        return useIncludeSyntax;
    }
    
    
    public static void log(String text) {
        System.out.println("Log: "+text);
    }

    public static void main(String[] args) {
        
        String rawText=
            "xxx\n" + 
            "+#EVT_ERR_HUMAN_TASK:+# " + 
            "xxx";

        CodeGenerator generator=new CodeGenerator();
        generator.setAddVariablesAsPlaceHolder(false);
        generator.setUseIncludeSyntax(false);
        String convertedText=generator.generatePLSQLCode(rawText);
        System.out.println(convertedText);                                                     
                                                              
    }
}